@@ -14,7 +14,7 @@ from coverage.exceptions import NotPython
 from coverage.parser import PythonParser
 
 from tests.coveragetest import CoverageTest
-from tests.helpers import arcz_to_arcs, xfail_pypy38
+from tests.helpers import arcz_to_arcs
 
 
 class PythonParserTestBase(CoverageTest):
@@ -50,22 +50,6 @@ class PythonParserTest(PythonParserTestBase):
             2:1, 3:1, 4:2, 5:1, 7:1, 9:1, 10:1,
         }
 
-    def test_generator_exit_counts(self) -> None:
-        # https://github.com/nedbat/coveragepy/issues/324
-        parser = self.parse_text("""\
-            def gen(input):
-                for n in inp:
-                    yield (i * 2 for i in range(n))
-
-            list(gen([1,2,3]))
-            """)
-        assert parser.exit_counts() == {
-            1:1,    # def -> list
-            2:2,    # for -> yield; for -> exit
-            3:2,    # yield -> for;  genexp exit
-            5:1,    # list -> exit
-        }
-
     def test_try_except(self) -> None:
         parser = self.parse_text("""\
             try:
@@ -79,7 +63,7 @@ class PythonParserTest(PythonParserTestBase):
             b = 9
             """)
         assert parser.exit_counts() == {
-            1: 1, 2:1, 3:2, 4:1, 5:2, 6:1, 7:1, 8:1, 9:1,
+            1: 1, 2:1, 3:1, 4:1, 5:1, 6:1, 7:1, 8:1, 9:1,
         }
 
     def test_excluded_classes(self) -> None:
@@ -92,9 +76,7 @@ class PythonParserTest(PythonParserTestBase):
                 class Bar:
                     pass
             """)
-        assert parser.exit_counts() == {
-            1:0, 2:1, 3:1,
-        }
+        assert parser.exit_counts() == { 2:1, 3:1 }
 
     def test_missing_branch_to_excluded_code(self) -> None:
         parser = self.parse_text("""\
@@ -124,30 +106,23 @@ class PythonParserTest(PythonParserTestBase):
             """)
         assert parser.exit_counts() == { 1:1, 2:1, 3:1, 6:1 }
 
-    def test_indentation_error(self) -> None:
-        msg = (
-            "Couldn't parse '<code>' as Python source: " +
-            "'unindent does not match any outer indentation level.*' at line 3"
+    @pytest.mark.parametrize("text", [
+        pytest.param("0 spaces\n  2\n 1", id="bad_indent"),
+        pytest.param("'''", id="string_eof"),
+        pytest.param("$hello", id="dollar"),
+        # on 3.10 this passes ast.parse but fails on tokenize.generate_tokens
+        pytest.param(
+            "\r'\\\n'''",
+            id="leading_newline_eof",
+            marks=[
+                pytest.mark.skipif(env.PYVERSION >= (3, 12), reason="parses fine in 3.12"),
+            ]
         )
+    ])
+    def test_not_python(self, text: str) -> None:
+        msg = r"Couldn't parse '<code>' as Python source: '.*' at line \d+"
         with pytest.raises(NotPython, match=msg):
-            _ = self.parse_text("""\
-                0 spaces
-                  2
-                 1
-                """)
-
-    def test_token_error(self) -> None:
-        submsgs = [
-            r"EOF in multi-line string",                                        # before 3.12.0b1
-            r"unterminated triple-quoted string literal .detected at line 1.",  # after 3.12.0b1
-        ]
-        msg = (
-            r"Couldn't parse '<code>' as Python source: '"
-            + r"(" + "|".join(submsgs) + ")"
-            + r"' at line 1"
-        )
-        with pytest.raises(NotPython, match=msg):
-            _ = self.parse_text("'''")
+            _ = self.parse_text(text)
 
     def test_empty_decorated_function(self) -> None:
         parser = self.parse_text("""\
@@ -164,22 +139,50 @@ class PythonParserTest(PythonParserTestBase):
             """)
 
         expected_statements = {1, 2, 4, 5, 8, 9, 10}
-        expected_arcs = set(arcz_to_arcs(".1 14 45 58 89 9.  .2 2.  -8A A-8"))
+        expected_arcs = set(arcz_to_arcs("14 45 58 89 9.  2.  A-8"))
         expected_exits = {1: 1, 2: 1, 4: 1, 5: 1, 8: 1, 9: 1, 10: 1}
 
         if env.PYBEHAVIOR.docstring_only_function:
             # 3.7 changed how functions with only docstrings are numbered.
-            expected_arcs.update(set(arcz_to_arcs("-46 6-4")))
+            expected_arcs.update(set(arcz_to_arcs("6-4")))
             expected_exits.update({6: 1})
 
-        if env.PYBEHAVIOR.trace_decorator_line_again:
-            expected_arcs.update(set(arcz_to_arcs("54 98")))
-            expected_exits.update({9: 2, 5: 2})
-
         assert expected_statements == parser.statements
         assert expected_arcs == parser.arcs()
         assert expected_exits == parser.exit_counts()
 
+    def test_nested_context_managers(self) -> None:
+        # https://github.com/nedbat/coveragepy/issues/1876
+        parser = self.parse_text("""\
+            a = 1
+            with suppress(ValueError):
+                with suppress(ValueError):
+                    x = 4
+                with suppress(ValueError):
+                    x = 6
+                with suppress(ValueError):
+                    x = 8
+            a = 9
+            """)
+
+        one_nine = set(range(1, 10))
+        assert parser.statements == one_nine
+        assert parser.exit_counts() == dict.fromkeys(one_nine, 1)
+
+    def test_module_docstrings(self) -> None:
+        parser = self.parse_text("""\
+            '''The docstring on line 1'''
+            a = 2
+            """)
+        assert {2} == parser.statements
+
+        parser = self.parse_text("""\
+            # Docstring is not line 1
+            '''The docstring on line 2'''
+            a = 3
+            """)
+        assert {3} == parser.statements
+
     def test_fuzzed_double_parse(self) -> None:
         # https://bugs.chromium.org/p/oss-fuzz/issues/detail?id=50381
         # The second parse used to raise `TypeError: 'NoneType' object is not iterable`
@@ -687,7 +690,6 @@ class ExclusionParserTest(PythonParserTestBase):
         )
         assert parser.statements == {1, 6}
 
-    @xfail_pypy38
     def test_decorator_pragmas(self) -> None:
         parser = self.parse_text("""\
             # 1
@@ -721,7 +723,6 @@ class ExclusionParserTest(PythonParserTestBase):
         assert parser.raw_statements == raw_statements
         assert parser.statements == {8}
 
-    @xfail_pypy38
     def test_decorator_pragmas_with_colons(self) -> None:
         # A colon in a decorator expression would confuse the parser,
         # ending the exclusion of the decorated function.
@@ -754,6 +755,217 @@ class ExclusionParserTest(PythonParserTestBase):
         assert parser.raw_statements == {1, 2, 3, 5, 6, 7, 8}
         assert parser.statements == {1, 2, 3}
 
+    def test_over_exclusion_bug1779(self) -> None:
+        # https://github.com/nedbat/coveragepy/issues/1779
+        parser = self.parse_text("""\
+            import abc
+
+            class MyProtocol:               # nocover 3
+                @abc.abstractmethod         # nocover 4
+                def my_method(self) -> int:
+                    ...     # 6
+
+            def function() -> int:
+                return 9
+            """)
+        assert parser.raw_statements == {1, 3, 4, 5, 6, 8, 9}
+        assert parser.statements == {1, 8, 9}
+
+    def test_multiline_exclusion_single_line(self) -> None:
+        regex = r"print\('.*'\)"
+        parser = self.parse_text("""\
+            def foo():
+                print('Hello, world!')
+            """, regex)
+        assert parser.lines_matching(regex) == {2}
+        assert parser.raw_statements == {1, 2}
+        assert parser.statements == {1}
+
+    def test_multiline_exclusion_suite(self) -> None:
+        # A multi-line exclusion that matches a colon line still excludes the entire block.
+        regex = r"if T:\n\s+print\('Hello, world!'\)"
+        parser = self.parse_text("""\
+            def foo():
+                if T:
+                    print('Hello, world!')
+                    print('This is a multiline regex test.')
+            a = 5
+            """, regex)
+        assert parser.lines_matching(regex) == {2, 3}
+        assert parser.raw_statements == {1, 2, 3, 4, 5}
+        assert parser.statements == {1, 5}
+
+    def test_multiline_exclusion_no_match(self) -> None:
+        regex = r"nonexistent"
+        parser = self.parse_text("""\
+            def foo():
+                print('Hello, world!')
+            """, regex)
+        assert parser.lines_matching(regex) == set()
+        assert parser.raw_statements == {1, 2}
+        assert parser.statements == {1, 2}
+
+    def test_multiline_exclusion_no_source(self) -> None:
+        regex = r"anything"
+        parser = PythonParser(text="", filename="dummy.py", exclude=regex)
+        assert parser.lines_matching(regex) == set()
+        assert parser.raw_statements == set()
+        assert parser.statements == set()
+
+    def test_multiline_exclusion_all_lines_must_match(self) -> None:
+        # https://github.com/nedbat/coveragepy/issues/996
+        regex = r"except ValueError:\n\s*print\('false'\)"
+        parser = self.parse_text("""\
+            try:
+                a = 2
+                print('false')
+            except ValueError:
+                print('false')
+            except ValueError:
+                print('something else')
+            except IndexError:
+                print('false')
+            """, regex)
+        assert parser.lines_matching(regex) == {4, 5}
+        assert parser.raw_statements == {1, 2, 3, 4, 5, 6, 7, 8, 9}
+        assert parser.statements == {1, 2, 3, 6, 7, 8, 9}
+
+    def test_multiline_exclusion_multiple_matches(self) -> None:
+        regex = r"print\('.*'\)\n\s+. = \d"
+        parser = self.parse_text("""\
+            def foo():
+                print('Hello, world!')
+                a = 5
+            def bar():
+                print('Hello again!')
+                b = 6
+            """, regex)
+        assert parser.lines_matching(regex) == {2, 3, 5, 6}
+        assert parser.raw_statements == {1, 2, 3, 4, 5, 6}
+        assert parser.statements == {1, 4}
+
+    def test_multiline_exclusion_suite2(self) -> None:
+        regex = r"print\('Hello, world!'\)\n\s+if T:"
+        parser = self.parse_text("""\
+            def foo():
+                print('Hello, world!')
+                if T:
+                    print('This is a test.')
+            """, regex)
+        assert parser.lines_matching(regex) == {2, 3}
+        assert parser.raw_statements == {1, 2, 3, 4}
+        assert parser.statements == {1}
+
+    def test_multiline_exclusion_match_all(self) -> None:
+        regex = (
+            r"def foo\(\):\n\s+print\('Hello, world!'\)\n"
+            + r"\s+if T:\n\s+print\('This is a test\.'\)"
+        )
+        parser = self.parse_text("""\
+            def foo():
+                print('Hello, world!')
+                if T:
+                    print('This is a test.')
+            """, regex)
+        assert parser.lines_matching(regex) == {1, 2, 3, 4}
+        assert parser.raw_statements == {1, 2, 3, 4}
+        assert parser.statements == set()
+
+    def test_multiline_exclusion_block(self) -> None:
+        # https://github.com/nedbat/coveragepy/issues/1803
+        regex = "# no cover: start(?s:.)*?# no cover: stop"
+        parser = self.parse_text("""\
+            a = my_function1()
+            if debug:
+                msg = "blah blah"
+                # no cover: start
+                log_message(msg, a)
+                b = my_function2()
+                # no cover: stop
+            """, regex)
+        assert parser.lines_matching(regex) == {4, 5, 6, 7}
+        assert parser.raw_statements == {1, 2, 3, 5, 6}
+        assert parser.statements == {1, 2, 3}
+
+    @pytest.mark.skipif(not env.PYBEHAVIOR.match_case, reason="Match-case is new in 3.10")
+    def test_multiline_exclusion_block2(self) -> None:
+        # https://github.com/nedbat/coveragepy/issues/1797
+        regex = r"case _:\n\s+assert_never\("
+        parser = self.parse_text("""\
+            match something:
+                case type_1():
+                    logic_1()
+                case type_2():
+                    logic_2()
+                case _:
+                    assert_never(something)
+            match something:
+                case type_1():
+                    logic_1()
+                case type_2():
+                    logic_2()
+                case _:
+                    print("Default case")
+            """, regex)
+        assert parser.lines_matching(regex) == {6, 7}
+        assert parser.raw_statements == {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14}
+        assert parser.statements == {1, 2, 3, 4, 5, 8, 9, 10, 11, 12, 13, 14}
+
+    def test_multiline_exclusion_block3(self) -> None:
+        # https://github.com/nedbat/coveragepy/issues/1741
+        # This will only work if there's exactly one return statement in the rest of the function
+        regex = r"# no cover: to return(?s:.)*?return"
+        parser = self.parse_text("""\
+            def my_function(args, j):
+                if args.command == Commands.CMD.value:
+                    return cmd_handler(j, args)
+                # no cover: to return
+                print(f"Command '{args.command}' was not handled.", file=sys.stderr)
+                parser.print_help(file=sys.stderr)
+
+                return os.EX_USAGE
+            print("not excluded")
+            """, regex)
+        assert parser.lines_matching(regex) == {4, 5, 6, 7, 8}
+        assert parser.raw_statements == {1, 2, 3, 5, 6, 8, 9}
+        assert parser.statements == {1, 2, 3, 9}
+
+    def test_multiline_exclusion_whole_source(self) -> None:
+        # https://github.com/nedbat/coveragepy/issues/118
+        regex = r"\A(?s:.*# pragma: exclude file.*)\Z"
+        parser = self.parse_text("""\
+            import coverage
+            # pragma: exclude file
+            def the_void() -> None:
+                if "py" not in __file__:
+                    print("Not a Python file.")
+                print("Everything here is excluded.")
+
+                return
+            print("Excluded too")
+            """, regex)
+        assert parser.lines_matching(regex) == {1, 2, 3, 4, 5, 6, 7, 8, 9, 10}
+        assert parser.raw_statements == {1, 3, 4, 5, 6, 8, 9}
+        assert parser.statements == set()
+
+    def test_multiline_exclusion_from_marker(self) -> None:
+        # https://github.com/nedbat/coveragepy/issues/118
+        regex = r"# pragma: rest of file(?s:.)*\Z"
+        parser = self.parse_text("""\
+            import coverage
+            # pragma: rest of file
+            def the_void() -> None:
+                if "py" not in __file__:
+                    print("Not a Python file.")
+                print("Everything here is excluded.")
+
+                return
+            print("Excluded too")
+            """, regex)
+        assert parser.lines_matching(regex) == {2, 3, 4, 5, 6, 7, 8, 9, 10}
+        assert parser.raw_statements == {1, 3, 4, 5, 6, 8, 9}
+        assert parser.statements == {1}
+
 
 class ParserMissingArcDescriptionTest(PythonParserTestBase):
     """Tests for PythonParser.missing_arc_description."""
@@ -775,45 +987,25 @@ class ParserMissingArcDescriptionTest(PythonParserTestBase):
                     thing(12)
                 more_stuff(13)
             """)
-        expected = "line 1 didn't jump to line 2, because the condition on line 1 was never true"
+        expected = "line 1 didn't jump to line 2 because the condition on line 1 was never true"
         assert expected == parser.missing_arc_description(1, 2)
-        expected = "line 1 didn't jump to line 3, because the condition on line 1 was always true"
+        expected = "line 1 didn't jump to line 3 because the condition on line 1 was always true"
         assert expected == parser.missing_arc_description(1, 3)
         expected = (
-            "line 6 didn't return from function 'func5', " +
+            "line 6 didn't return from function 'func5' " +
             "because the loop on line 6 didn't complete"
         )
         assert expected == parser.missing_arc_description(6, -5)
-        expected = "line 6 didn't jump to line 7, because the loop on line 6 never started"
+        expected = "line 6 didn't jump to line 7 because the loop on line 6 never started"
         assert expected == parser.missing_arc_description(6, 7)
-        expected = "line 11 didn't jump to line 12, because the condition on line 11 was never true"
+        expected = "line 11 didn't jump to line 12 because the condition on line 11 was never true"
         assert expected == parser.missing_arc_description(11, 12)
         expected = (
-            "line 11 didn't jump to line 13, " +
+            "line 11 didn't jump to line 13 " +
             "because the condition on line 11 was always true"
         )
         assert expected == parser.missing_arc_description(11, 13)
 
-    def test_missing_arc_descriptions_for_small_callables(self) -> None:
-        parser = self.parse_text("""\
-            callables = [
-                lambda: 2,
-                (x for x in range(3)),
-                {x:1 for x in range(4)},
-                {x for x in range(5)},
-            ]
-            x = 7
-            """)
-        expected = "line 2 didn't finish the lambda on line 2"
-        assert expected == parser.missing_arc_description(2, -2)
-        expected = "line 3 didn't finish the generator expression on line 3"
-        assert expected == parser.missing_arc_description(3, -3)
-        if env.PYBEHAVIOR.comprehensions_are_functions:
-            expected = "line 4 didn't finish the dictionary comprehension on line 4"
-            assert expected == parser.missing_arc_description(4, -4)
-            expected = "line 5 didn't finish the set comprehension on line 5"
-            assert expected == parser.missing_arc_description(5, -5)
-
     def test_missing_arc_descriptions_for_exceptions(self) -> None:
         parser = self.parse_text("""\
             try:
@@ -824,102 +1016,21 @@ class ParserMissingArcDescriptionTest(PythonParserTestBase):
                 print("yikes")
             """)
         expected = (
-            "line 3 didn't jump to line 4, " +
+            "line 3 didn't jump to line 4 " +
             "because the exception caught by line 3 didn't happen"
         )
         assert expected == parser.missing_arc_description(3, 4)
         expected = (
-            "line 5 didn't jump to line 6, " +
+            "line 5 didn't jump to line 6 " +
             "because the exception caught by line 5 didn't happen"
         )
         assert expected == parser.missing_arc_description(5, 6)
 
-    def test_missing_arc_descriptions_for_finally(self) -> None:
-        parser = self.parse_text("""\
-            def function():
-                for i in range(2):
-                    try:
-                        if something(4):
-                            break
-                        elif something(6):
-                            x = 7
-                        else:
-                            if something(9):
-                                continue
-                            else:
-                                continue
-                        if also_this(13):
-                            return 14
-                        else:
-                            raise Exception(16)
-                    finally:
-                        this_thing(18)
-                that_thing(19)
-            """)
-        if env.PYBEHAVIOR.finally_jumps_back:
-            expected = "line 18 didn't jump to line 5, because the break on line 5 wasn't executed"
-            assert expected == parser.missing_arc_description(18, 5)
-            expected = "line 5 didn't jump to line 19, because the break on line 5 wasn't executed"
-            assert expected == parser.missing_arc_description(5, 19)
-            expected = (
-                "line 18 didn't jump to line 10, " +
-                "because the continue on line 10 wasn't executed"
-            )
-            assert expected == parser.missing_arc_description(18, 10)
-            expected = (
-                "line 10 didn't jump to line 2, " +
-                "because the continue on line 10 wasn't executed"
-            )
-            assert expected == parser.missing_arc_description(10, 2)
-            expected = (
-                "line 18 didn't jump to line 14, " +
-                "because the return on line 14 wasn't executed"
-            )
-            assert expected == parser.missing_arc_description(18, 14)
-            expected = (
-                "line 14 didn't return from function 'function', " +
-                "because the return on line 14 wasn't executed"
-            )
-            assert expected == parser.missing_arc_description(14, -1)
-            expected = (
-                "line 18 didn't except from function 'function', " +
-                "because the raise on line 16 wasn't executed"
-            )
-            assert expected == parser.missing_arc_description(18, -1)
-        else:
-            expected = (
-                "line 18 didn't jump to line 19, " +
-                "because the break on line 5 wasn't executed"
-            )
-            assert expected == parser.missing_arc_description(18, 19)
-            expected = (
-                "line 18 didn't jump to line 2, " +
-                    "because the continue on line 10 wasn't executed" +
-                " or " +
-                    "the continue on line 12 wasn't executed"
-            )
-            assert expected == parser.missing_arc_description(18, 2)
-            expected = (
-                "line 18 didn't except from function 'function', " +
-                    "because the raise on line 16 wasn't executed" +
-                " or " +
-                "line 18 didn't return from function 'function', " +
-                    "because the return on line 14 wasn't executed"
-            )
-            assert expected == parser.missing_arc_description(18, -1)
 
-    def test_missing_arc_descriptions_bug460(self) -> None:
-        parser = self.parse_text("""\
-            x = 1
-            d = {
-                3: lambda: [],
-                4: lambda: [],
-            }
-            x = 6
-            """)
-        assert parser.missing_arc_description(2, -3) == "line 3 didn't finish the lambda on line 3"
+@pytest.mark.skipif(not env.PYBEHAVIOR.match_case, reason="Match-case is new in 3.10")
+class MatchCaseMissingArcDescriptionTest(PythonParserTestBase):
+    """Missing arc descriptions for match/case."""
 
-    @pytest.mark.skipif(not env.PYBEHAVIOR.match_case, reason="Match-case is new in 3.10")
     def test_match_case(self) -> None:
         parser = self.parse_text("""\
             match command.split():
@@ -930,15 +1041,33 @@ class ParserMissingArcDescriptionTest(PythonParserTestBase):
             print(match)                                            # 6
             """)
         assert parser.missing_arc_description(2, 3) == (
-            "line 2 didn't jump to line 3, because the pattern on line 2 never matched"
+            "line 2 didn't jump to line 3 because the pattern on line 2 never matched"
         )
         assert parser.missing_arc_description(2, 4) == (
-            "line 2 didn't jump to line 4, because the pattern on line 2 always matched"
+            "line 2 didn't jump to line 4 because the pattern on line 2 always matched"
         )
         assert parser.missing_arc_description(4, 6) == (
-            "line 4 didn't jump to line 6, because the pattern on line 4 always matched"
+            "line 4 didn't jump to line 6 because the pattern on line 4 always matched"
         )
 
+    def test_final_wildcard(self) -> None:
+        parser = self.parse_text("""\
+            match command.split():
+                case ["go", direction] if direction in "nesw":      # 2
+                    match = f"go: {direction}"
+                case _:                                             # 4
+                    match = "no go"
+            print(match)                                            # 6
+            """)
+        assert parser.missing_arc_description(2, 3) == (
+            "line 2 didn't jump to line 3 because the pattern on line 2 never matched"
+        )
+        assert parser.missing_arc_description(2, 4) == (
+            "line 2 didn't jump to line 4 because the pattern on line 2 always matched"
+        )
+        # 4-6 isn't a possible arc, so the description is generic.
+        assert parser.missing_arc_description(4, 6) == "line 4 didn't jump to line 6"
+
 
 class ParserFileTest(CoverageTest):
     """Tests for coverage.py's code parsing from files."""
